# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for extending and configuring the logging framework.

This module is called ``logs`` instead of ``logging`` to avoid conflicts
with the builtin module. Since this module is a helper, it is expected
to be imported alongside the builtin module.
"""

__all__ = [
    'KEEP', 'KEEP_WARN', 'OVERWRITE', 'OVERWRITE_WARN', 'RAISE',
    'add_logging_level', 'add_trace_level', 'configure_logger',
    'reset_handlers',
    'LogMaxFilter', 'MetaLoggableType',
]


import abc
import logging
import sys
import warnings

from . import Sentinel


#: Default format string for the root logger. This string is set up by
#: the :py:func:`configure_logger` method.
_log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'

#: When adding a new logging level, with :py:func:`add_logging_level`,
#: silently keep the old level in case of conflict.
KEEP = 'keep'


#: When adding a new logging level, with :py:func:`add_logging_level`,
#: keep the old level in case of conflict, and issue a warning.
KEEP_WARN = 'keep-warn'


#: When adding a new logging level, with :py:func:`add_logging_level`,
#: silently overwrite any existing level in case of conflict.
OVERWRITE = 'overwrite'


#: When adding a new logging level, with :py:func:`add_logging_level`,
#: overwrite any existing level in case of conflict, and issue a
#: warning.
OVERWRITE_WARN = 'overwrite-warn'


#: When adding a new logging level, with :py:func:`add_logging_level`,
#: raise an error in case of conflict.
RAISE = 'raise'


class MetaLoggableType(abc.ABCMeta):
    """
    A metaclass for assigning a logger with a properly named channel to
    classes.

    The logger channel will be the fully qualified name of the class
    including package and module prefixes.

    .. py:attribute:: __namespace__

       If this attribute is found in the class definition, it will be
       prefixed to the qualified name (with a dot).

    .. py:attribute:: logger

       This attribute is assigned to all new classes based on the name
       and possibly :py:attr:`__namespace__`.
    """

    def __init__(cls, name, bases, dct):
        """
        Intialize a newly created class with a :py:attr:`logger`
        attribute.

        If a `__namespace__` attribute is found in the class, its contents is
        prefixed to the logger name.
        """
        name = cls.__module__ + '.' + cls.__qualname__
        if '__namespace__' in dct:
            name = dct['__namespace__'] + '.' + name
        cls.logger = logging.getLogger(name)
        dct['logger'] = cls.logger
        return super().__init__(name, bases, dct)


def add_logging_level(level_name, level_num, method_name=None,
                      if_exists=KEEP, *, exc_info=False, stack_info=False):
    """
    Comprehensively add a new logging level to the :py:mod:`logging`
    module and the currently configured logging class.

    The `if_exists` parameter determines the behavior if the level
    name is already an attribute of the :py:mod:`logging` module or if
    the method name is already present, unless the attributes are
    configured to the exact values requested. Partial registration is
    considered a conflict. Even a complete registration will be
    overwritten if ``if_exists in (OVERWRITE, OVERWRITE_WARN)`` (without
    a warning of course).

    This function also accepts alternate default values for the keyword
    arguments ``exc_info`` and ``stack_info`` that are optional for
    every logging method. Setting alternate defaults allows levels for
    which exceptions or stacks are always logged.

    Parameters
    ----------
    level_name : str
        Becomes an attribute of the :py:mod:`logging` module with the
        value ``level_num``.
    level_num : int
        The numerical value of the new level.
    method_name : str
        The name of the convenience method for both :py:mod:`logging`
        itself and the class returned by
        :py:func:`logging.getLoggerClass` (usually just
        :py:class:`logging.Logger`). If ``method_name`` is not
        specified, ``level_name.lower()`` is used instead.
    if_exists : {KEEP, KEEP_WARN, OVERWRITE, OVERWRITE_WARN, RAISE}
        What to do if a level with `level_name` appears to already be
        registered in the :py:mod:`logging` module:

        :py:const:`KEEP`
            Silently keep the old level as-is.
        :py:const:`KEEP_WARN`
            Keep the old level around and issue a warning.
        :py:const:`OVERWRITE`
            Silently overwrite the old level.
        :py:const:`OVERWRITE_WARN`
            Overwrite the old level and issue a warning.
        :py:const:`RAISE`
            Raise an error.

        The default is :py:const:`KEEP_WARN`.
    exc_info : bool
        Default value for the ``exc_info`` parameter of the new method.
    stack_info : bool
        Default value for the ``stack_info`` parameter of the new
        method.

    Examples
    --------
    >>> add_logging_level('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    >>> add_logging_level('XTRACE', 2, exc_info=True)
    >>> logging.getLogger(__name__).setLevel(logging.XTRACE)
    >>> try:
    >>>     1 / 0
    >>> except:
    >>>     # This line will log the exception
    >>>     logging.getLogger(__name__).xtrace('that failed')
    >>>     # This one will not
    >>>     logging.xtrace('so did this', exc_info=False)

    The ``TRACE`` level can be added using :py:func:`add_trace_level`.

    Note
    ----
    Before adding new levels, please see the cautionary note here:
    https://docs.python.org/3/howto/logging.html#custom-levels.
    """
    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def for_logger_adapter(self, msg, *args, **kwargs):
        kwargs.setdefault('exc_info', exc_info)
        kwargs.setdefault('stack_info', stack_info)
        kwargs.setdefault('stacklevel', 2)
        self.log(level_num, msg, *args, **kwargs)

    def for_logger_class(self, msg, *args, **kwargs):
        if self.isEnabledFor(level_num):
            kwargs.setdefault('exc_info', exc_info)
            kwargs.setdefault('stack_info', stack_info)
            kwargs.setdefault('stacklevel', 2)
            self._log(level_num, msg, args, **kwargs)

    def for_logging_module(*args, **kwargs):
        kwargs.setdefault('exc_info', exc_info)
        kwargs.setdefault('stack_info', stack_info)
        kwargs.setdefault('stacklevel', 2)
        logging.log(level_num, *args, **kwargs)

    if not method_name:
        method_name = level_name.lower()
    if method_name == level_name:
        raise ValueError('Method name must differ from level name')

    # The number of items required for a full registration is 5
    items_found = 0
    # Items that are found complete but are not expected values
    items_conflict = 0

    def check_conflict(conflict, message):
        if conflict and if_exists == RAISE:
            raise AttributeError(message)
        return conflict

    def check_func_conflict(func, name, original_name, is_func, target):
        conflict = not (
            callable(func) and
            getattr(func, '_original_name', None) == original_name and
            getattr(func, '_exc_info', None) == exc_info and
            getattr(func, '_stack_info', None) == stack_info
        )
        return check_conflict(
            conflict, '{} {!r} already defined in {}'.format(
                "Function" if is_func else "Method", name, target
            )
        )
        return conflict

    # Lock because logger class and level name are queried and set
    logging._acquireLock()
    try:
        registered_num = logging.getLevelName(level_name)
        logger_class = logging.getLoggerClass()
        logger_adapter = logging.LoggerAdapter

        if registered_num != 'Level ' + level_name:
            items_found += 1
            items_conflict += check_conflict(
                registered_num != level_num, 'Level {!r} already registered '
                'in logging module'.format(level_name)
            )

        current_level = getattr(logging, level_name, Sentinel)
        if current_level is not Sentinel:
            items_found += 1
            items_conflict += check_conflict(
                current_level != level_num, 'Level {!r} already defined '
                'in logging module'.format(level_name)
            )

        logging_func = getattr(logging, method_name, Sentinel)
        if logging_func is not Sentinel:
            items_found += 1
            items_conflict += check_func_conflict(
                logging_func, method_name, for_logging_module.__name__,
                True, 'logging module'
            )

        logger_method = getattr(logger_class, method_name, Sentinel)
        if logger_method is not Sentinel:
            items_found += 1
            items_conflict += check_func_conflict(
                logger_method, method_name, for_logger_class.__name__,
                False, 'logger class'
            )

        adapter_method = getattr(logger_adapter, method_name, Sentinel)
        if adapter_method is not Sentinel:
            items_found += 1
            items_conflict += check_func_conflict(
                adapter_method, method_name, for_logger_adapter.__name__,
                False, 'logger adapter'
            )

        if items_found > 0:
            # items_found >= items_conflict always
            if (items_conflict or items_found < 5) and \
                    if_exists in (KEEP_WARN, OVERWRITE_WARN):
                action = 'Keeping' if if_exists == KEEP_WARN else 'Overwriting'
                if items_conflict:
                    problem = 'has conflicting definition'
                    items = items_conflict
                else:
                    problem = 'is partially configured'
                    items = items_found
                warnings.warn(
                    'Logging level {!r} {} already ({}/5 items): {}'.format(
                        level_name, problem, items, action)
                )

            if if_exists in (KEEP, KEEP_WARN):
                return

        # Make sure the method names are set to sensible values, but
        # preserve the names of the old methods for future verification.
        def label_func(func):
            func._original_name = func.__name__
            func.__name__ = method_name
            func._exc_info = exc_info
            func._stack_info = stack_info
        label_func(for_logging_module)
        label_func(for_logger_class)
        label_func(for_logger_adapter)

        # Actually add the new level
        logging.addLevelName(level_num, level_name)
        setattr(logging, level_name, level_num)
        setattr(logging, method_name, for_logging_module)
        setattr(logger_class, method_name, for_logger_class)
        setattr(logger_adapter, method_name, for_logger_adapter)
    finally:
        logging._releaseLock()


def add_trace_level(if_exists=KEEP_WARN):
    """
    Add a new ``TRACE`` level to the :py:mod:`logging` module.

    The numerical trace level is ``5`` lower than
    :py:data:`~logging.DEBUG`. It does not log stack or exception
    information by default. A ``trace`` method will be added to the
    :py:mod:`logging` module, to the current default
    :py:class:`~logging.Logger` class, and to the
    :py:class:`~logging.LoggerAdapter` class.
    """
    add_logging_level('TRACE', logging.DEBUG - 5, 'trace', if_exists)


def add_detail_level(if_exists=KEEP_WARN):
    """
    Add a new ``DETAIL`` level to the :py:mod:`logging` module.

    The numerical trace level is ``5`` higher than
    :py:data:`~logging.DEBUG`. It does not log stack or exception
    information by default. A ``detail`` method will be added to the
    :py:mod:`logging` module, to the current default
    :py:class:`~logging.Logger` class, and to the
    :py:class:`~logging.LoggerAdapter` class.
    """
    add_logging_level('DETAIL', logging.DEBUG + 5, 'detail', if_exists)


def LogMaxFilter(level, inclusive=True):
    """
    Create a level-based filter that caps the maximum allowed log level.

    Levels can be compared either exclusively or inclusively to the
    threshold.

    Parameters
    ----------
    level : int
        The cutoff level: only messages below this will be passed
        through.
    inclusive : bool
        If True, messages at level will be cut off. Otherwise, only
        messages strictly more severe than `level` will be cut off.

    Return
    ------
    filter :
        A callable filter that operates on log records.

    Notes
    -----
    This function returns a callable rather than an object with a
    ``filter`` method, so it is not compatible with `logging` before
    Python 3.2.
    """
    if not isinstance(level, int):
        level = logging.getLevelName(level)
    if not isinstance(level, int):
        raise TypeError('Numerical level or '
                        'level that maps to number required')
    comparator = level.__gt__ if inclusive else level.__ge__

    def filter(log_record):
        return comparator(log_record.levelno)

    return filter


def _get_formatter(format=None):
    """
    Retrieve a consistent :py:class:`~logging.Formatter` object based
    on the input format.

    If `format` is already a :py:class:`~logging.Formatter`, return it
    as-is. If :py:obj:`None`, use a default format string. Otherwise, it
    is expected to be a string that initializes a proper
    :py:class:`~logging.Formatter` instance.
    """
    if isinstance(format, logging.Formatter):
        return format
    if format is None:
        format = _log_format
    return logging.Formatter(format)


def configure_logger(log_file=None, file_level='NOTSET',
                     log_stderr=True, stderr_level='WARNING',
                     log_stdout=False, stdout_level='INFO',
                     format_string=None, trace_warnings=True):
    """
    Set up the root logger based on the input parameters.

    A ``TRACE`` level is added to the :py:mod:`logging` module. The
    system-level automatic exception handler is set up to log uncaught
    errors. Warnings will always be captured by the logger, with
    optional tracebacks being logged by default.

    Parameters
    ----------
    log_file : None or str
        If not :py:obj:`None`, messages with level greater than or equal
        to `file_level` will go to the specified file.
    file_level : str
        The name of the minimum logging level that will be written to
        the file log if `log_file` is set. Defaults to ``'NOTSET'``.
        Case insensitive.
    log_stderr : bool
        If :py:obj:`True`, messages with level greater than or equal to
        `stderr_level` will be output to standard error. Defaults to
        :py:obj:`True`.
    stderr_level : str
        The name of the minimum logging level that will be output to
        standard error if `log_stderr` is set. Defaults to
        ``'WARNING'``.
    log_stdout : bool
        If :py:obj:`True`, messages with level greater than or equal to
        `stdout_level` will be output to standard output. Defaults to
        :py:obj:`False`. If `log_stderr` is set as well, only levels
        strictly less than `stderr_level` will be printed to standard
        output.
    stdout_level : str
        The name of the minimum logging level that will be output to
        standard error if `log_stdout` is set. Defaults to ``'INFO'``.
    format_string : str
        The log format. A missing (:py:obj:`None`) `format_string`
        defaults to
        ``'%(asctime)s - %(name)s - %(levelname)s - %(message)s'``. 
    trace_warnings : bool
        Whether or not to print tracebacks for actual warnings (not log
        entries with a warning level) caught by the Python global
        warning logger. Defaults to :py:obj:`True`. Custom
        :py:meth:`~logging.Logger.warning` methods are hooked into the
        logger for ``"py.warnings"``.
    """
    add_trace_level()

    # System-level logging hook courtesy of Stack Overflow post
    # http://stackoverflow.com/a/16993115/2988730
    def exception_handler(*args):
        """
        An exception hook that is meant to replace the default
        :py:func:`sys.excepthook`. Logs all uncaught exceptions to the
        root logger except for :py:exc:`KeyboardInterrupt`, which is
        passed directly to the default system hook.

        An additional hook can be called after logging the exception
        """
        if isinstance(args[0], KeyboardInterrupt):
            sys.__excepthook__(*args)
        else:
            root.critical("Uncaught exception", exc_info=args)

    sys.excepthook = exception_handler

    if trace_warnings:
        def warning_logger(self, msg, *args, **kwargs):
            """
            A replacement for the :py:meth:`~logging.Logger.warning`
            method that is set for the ``"py.warnings"`` logger to show
            a stack trace for logged warnings. Note that warnings logged
            explicitly with ``logger.log(logging.WARNING, ...)`` will
            not pass through this handler.
            """
            if self.isEnabledFor(logging.WARNING):
                kwargs.setdefault('stack_info', True)
                self._log(logging.WARNING, msg, args, **kwargs)

        logger = logging.getLogger("py.warnings")

        # Bind methods according to http://stackoverflow.com/a/1015405/2988730
        logger.warning = warning_logger.__get__(logger, type(logger))
        logger.warn = logger.warning  # For good measure

    logging.captureWarnings(True)

    # Remaining logger setup courtesy of Stack Overflow post
    # http://stackoverflow.com/a/24978464/2988730
    formatter = _get_formatter(format_string)

    root = logging.getLogger()
    # Apparently, stdout won't show up without this...
    root.setLevel(logging.NOTSET)

    if log_file:
        reset_handlers(logging.FileHandler(log_file, mode='w'),
                       level=file_level, format=formatter, logger=root)
    if log_stderr:
        reset_handlers(logging.StreamHandler(sys.stderr), level=stderr_level,
                       format=formatter, logger=root, filter_type=True,
                       filter_hook=lambda h: h.stream == sys.stderr)
    if log_stdout:
        stdout_handler = logging.StreamHandler(sys.stdout)
        reset_handlers(stdout_handler, level=stdout_level, format=formatter,
                       logger=root, filter_type=True,
                       filter_hook=lambda h: h.stream == sys.stdout)
        if log_stderr:
            stdout_handler.addFilter(LogMaxFilter(
                    logging.getLevelName(stderr_level), False))


def reset_handlers(handler, level='NOTSET', format=None, logger=None,
                   filter_type=None, filter_hook=None, remove_hook=None):
    """
    Remove all handlers of a given class from `logger` (root by
    default), and replaces them with `handler`.

    If a handler that is being removed has a ``close`` method, it will
    be called, unless `remove_hook` is explicitly set.

    If both `filter_type` and `filter_hook` are set, both conditions
    must be met in order for a handler to be removed.

    Parameters
    ----------
    handler : logging.Handler
        The new handler to place in the list.
    level : str
        The case insensitive name of the minimum logging level to
        set for `handler`. Defaults to ``'NOTSET'``. This will not
        affect the level set for the logger.
    format : None or str or logging.Formatter
        Format for the log output strings.
    logger : None or logging.Logger
        The logger to set the handler for. Defaults to the root logger.
        Neither child nor ancestor loggers will be affected by this
        operation.
    filter_type : None, bool or type
        The type of objects to remove from the current list of handlers.
        If a superclass of `handler`, it will be used as the filter
        instead of ``type(handler)``. Any other type will raise an error.
        If :py:obj:`None`, then filtering by type will be done only if
        `filter_hook` is not set. A :py:class:`bool` explicitly sets
        filtering by ``type(handler)`` on and off regardless of
        `filter_hook`.
    filter_hook : None or callable
        A function that accepts a :py:class:`~logging.Handler` and
        returns a :py:class:`bool`. :py:obj:`True` indicates that an
        object should be removed from the list of handlers.
    remove_hook : None or callable
        A function that accepts a :py:class:`~logging.Handler` and
        performs some additional action such as closing it. The default
        behavior is to invoke ``close()`` on all handlers that are being
        removed if they have that method.
    """
    if filter_type is None:
        if filter_hook is None:
            filter_type = type(handler)
    elif filter_type is True:
        filter_type = type(handler)
    elif filter_type is False:
        filter_type = None
    else:
        if not isinstance(handler, filter_type):
            raise TypeError('{} is not a subclass of {}'.format(
                type(handler), filter_type))

    if filter_hook is None:
        if filter_type is None:
            def filter_func(h):
                return True
        else:
            def filter_func(h):
                return isinstance(h, filter_type)
    else:
        if filter_type is None:
            filter_func = filter_hook
        else:
            def filter_func(h):
                return isinstance(h, filter_type) and filter_hook(h)

    if remove_hook is None:
        def remove_hook(h):
            if hasattr(h, 'close') and callable(h.close):
                h.close()

    logging._acquireLock()
    try:
        if logger is None:
            logger = logging.getLogger()

        for h in list(logger.handlers):
            if filter_func(h):
                remove_hook(h)
                logger.removeHandler(h)

        handler.setLevel(logging.getLevelName(level))
        handler.setFormatter(_get_formatter(format))
        logger.addHandler(handler)
    finally:
        logging._releaseLock()



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Math utility functions that are otherwise uncategorized.
"""

__all__ = [
    'ang_diff_abs', 'ang_diff_min', 'ang_diff_pos', 'count_divisors',
    'ellipse', 'first_primes', 'full_width_half_max', 'primes_up_to',
    'real_divide', 'rms', 'round_sig', 'segment_distance', 'threshold'
]


import itertools
import math
import numpy
from scipy.stats import iqr

from .mapping import option_lookup


def round_sig(x, n):
    """
    Round the number `x` to `n` significant figures.

    Based on https://stackoverflow.com/a/3411435/2988730
    """
    if not x:
        return x
    digits = -math.floor(math.log10(abs(x))) + (n - 1)
    return round(x, digits)


def ellipse(*args, num_points=1e3, **kwargs):
    r"""
    .. py:function:: ellipse(a, [b=0,] c, d, e, f, *, numPoints=1e3)
       :noindex:
    .. py:function:: ellipse(a, b, h, k, *, theta=0, numPoints=1e3)
       :noindex:

    Return x-y arrays for an ellipse in one of two standard forms.

    The equations are either the quadratic form,

    .. math::

       ax^2 + bxy + cy^2 + dx + ey + f = 0

    or the simplified form,

    .. math::

       \frac{(x - h)^2}{a^2} + \frac{(y - k)^2}{b^2} = 1

    The first form allows for rotated ellipses through the :math:`bxy`
    coupling term. The second form allows it through the explicit
    (optional) angle parameter `theta`, specified in radians
    counterclockwise about ``(h, k)``.

    The number of points is specified by `numPoints`. Points are evenly
    distributed by angle, not by arc-length (unless the ellipse is a
    circle). The default number of points is 1000.

    Return
    ======
    x : numpy.ndarray
        Array of `num_points` x-coordinates.
    y : numpy.ndarray
        Array of `num_points` y-coordinates.

    Notes
    =====
    This code is loosely based on my Stack Overflow answer:
    http://stackoverflow.com/a/41231100/2988730, which is in turn
    loosely based on the forum post at
    http://www.sosmath.com/CBB/viewtopic.php?t=17029
    """
    t = numpy.linspace(0, 2 * math.pi, num_points)

    coeffCount = len(args)
    if 5 <= coeffCount <= 6:
        if coeffCount == 5:
            a, c, d, e, f = args
            b = 0
        if len(args) == 6:
            a, b, c, d, e, f = args
        if kwargs:
            raise ValueError('Only num_points can be a keyword '
                             'for quadratic form of ellipse')

        discriminant = b**2 - 4 * a * c
        if discriminant >= 0.0:
            name = 'parabloa' if discriminant == 0.0 else 'hyperbola'
            raise ValueError('Discriminant shows this to be '
                             'a {}, not an ellipse'.format(name))
        if b == 0:
            theta = 0
        else:
            theta = 0.5 * math.atan(b / (a - c))
            sin = math.sin(theta)
            cos = math.cos(theta)
            aa = a * cos**2 + b * sin * cos + c * sin**2
            cc = a * sin**2 - b * sin * cos + c * cos**2
            dd = d * cos + e * sin
            ee = -d * sin + e * cos
            a, b, c, d, e = aa, 0, cc, dd, ee
        ax_x = 1.0 / math.sqrt(a)
        ax_y = 1.0 / math.sqrt(c)
        scale = math.sqrt(d**2 / (4 * a) + e**2 / (4 * c) - f)
        h = -d / (2 * a)
        k = -e / (2 * c)
        x = xx = h + ax_x * scale * numpy.sin(t)
        y = yy = k + ax_y * scale * numpy.cos(t)
        if theta:
            x = xx * cos - yy * sin
            y = xx * sin + yy * cos
    elif coeffCount == 4:
        ax_x, ax_y, h, k = args
        x = xx = ax_x * numpy.sin(t)
        y = yy = ax_y * numpy.cos(t)
        theta = float(kwargs.pop('theta', 0.0))
        if theta:
            sin = math.sin(theta)
            cos = math.cos(theta)
            x = h + xx * cos - yy * sin
            y = k + xx * sin + yy * cos
        if kwargs:
           raise ValueError('Only num_points and theta can be a keyword '
                            'argument for simple form of ellipse')
    else:
        raise ValueError('Argument list must have 4 to 6 positional elements')

    return x, y


def full_width_half_max(x, y, factor=0.5, baseline=0.0, interp='linear', *,
                        return_points=False):
    """
    Compute the full-width, half-max metric for a dataset.

    The full-width half-max is the distance between where the data rises
    to half of the maximum for the last time before the max itself and
    where it falls below half of the maximum for the first time above
    the maximum.

    The actual fraction of the maximum that is used can be adjusted with
    `factor` (which defaults to 0.5). `factor` can be a number or the
    string ``'sigma'``, in which case it will be set to
    :math:`e^{-\\frac{1}{4}}`, the height of a Gaussian with unit
    amplitude one standard devition away from the mean. A fixed non-zero
    baseline can also be provided so that it does not have to be
    subtracted from `y` before being passed in.

    Since the actual `y` array is unlikely to contain the exact half-max
    value, an interpolation can be done. Currently, the following
    interpolation methods are supported:

      - ``'linear'``: linear interpolation betweent the x- and y-values
        surrouding the actual half-max point.
      - ``'nearest'``: nearest neighbor, i.e., use the x-value of the
        y-value that is closest to the half-max.

    Normally, a single scalar is returned. If `return_points` is
    :py:obj:`True`, however, two two-element tuples are returned as the
    second and third argument. Each tuple will contain an x-y pair of
    the intersection coordinates used to approximate the main return
    value. The first tuple will be for the left (rising) edge and the
    second will be for the right (trailing) edge.
    """
    imax = numpy.argmax(y)
    if factor == 'sigma':
        factor = math.exp(-0.25)
    # Also can be written as factor * y[imax] + (1.0 - factor) * baseline
    halfmax = factor * (y[imax] - baseline) + baseline
    # Always use <= to allow first and last y-value if they match exactly.
    rising = (y[:imax] <= halfmax) & (y[1:imax + 1] > halfmax)
    falling = (y[imax:-1] > halfmax) & (y[imax + 1:] <= halfmax)
    if not numpy.any(rising):
        raise ValueError(
            'left edge does not fall below {} of max'.format(factor)
        )
    if not numpy.any(falling):
        raise ValueError(
            'right edge does not fall below {} of max'.format(factor)
        )
    # Select last rising, first falling index.
    # Initial [0] because of nonzero's weird tuple return
    rising_index = numpy.nonzero(rising)[0][-1] + numpy.arange(2)
    falling_index = numpy.nonzero(falling)[0][0] + imax + numpy.arange(2)
    if interp == 'linear':
        def linterp(x0, x1, y0, y1, y):
            return x0 + (x1 - x0) * (y - y0) / (y1 - y0)
        rising = (linterp(*x[rising_index], *y[rising_index],halfmax),
                  halfmax)
        falling = (linterp(*x[falling_index], *y[falling_index], halfmax),
                   halfmax)
    elif interp == 'nearest':
        def nearest(ys, inds, y):
            return inds[numpy.argmin(numpy.abs(ys[inds] - y))]
        rising_i = nearest(y, rising_index, halfmax)
        falling_i = nearest(y, falling_index, halfmax)
        rising = (x[rising_i], y[rising_i])
        falling = (x[falling_i], y[falling_i])
    else:
        raise ValueError(
            'Invalid value of `interp` parameter. Expected '
            '{{"linear", "nearest"}}, found "{}".'.format(interp)
        )

    if return_points:
        return falling[0] - rising[0], rising, falling

    return falling[0] - rising[0]


def primes_up_to(n):
    """
    Generate a set containing all the primes less than or equal to `n`.

    `n` must be a number that represents an array size that can exist
    in memory. The implementation uses an extremely unoptimized version
    of the sieve of Eratosthenes.

    Parameters
    ----------
    n : int
        The largest number to generate primes up to (exclusive). If you
        want an inclusive range, add 1 to this input.

    Return
    ------
    primes : set
        A set of all the primes less than `n`.
    """
    sieve = numpy.arange(2, n, dtype=numpy.int_)
    for i in range(n - 2):
        p = sieve[i]
        if p:
            sieve[i+p::p] = 0
    return sieve[numpy.nonzero(sieve)]


def first_primes(n):
    """
    Generate a set with the first `n` prime numbers.

    This is a toy method that should probably not be used for large
    prime numbers. Instead of actively discarding all multiples of
    found primes, it checks new candidates against each element of the
    current set of primes.

    Parameters
    ----------
    n : int

    Return
    ------
    primes : set
    """
    primes = set()
    for i in itertools.count(2):
        if all(i % p for p in primes):
            primes.add(i)
        if len(primes) == n:
            break
    return primes


def count_divisors(n):
    """
    Counts the divisors of natural number `n`, including 1 and itself.

    For example, ``28`` has divisors ``1, 2, 4, 7, 14, 28``, so
    ``count_divisors(28) == 7``.
    """
    count = 0
    limit = int(math.sqrt(n))
    # Add one if the number is a square (limit won't be part of loop)
    extra = (limit**2 == n)
    for i in range(2, limit):
        if not n % i:
            count += 1
    # Add one to count because loop does not check if 1/n are factors
    return (count + 1) * 2 + extra


def real_divide(a, b, zero=0, out=None):
    """
    Divide real numbers, where the second may be zero.

    Parameters
    ----------
    a : array-like
        The divisor.
    b : array-like
        The dividend
    zero :
        The value to place in locations where `b` is zero.
    out : array-like or None
        An array of a suitable type and size to hold the result.
        If `None`, a new output array is allocated.

    Return
    ------
    numpy.ndarray :
        The result of applying :py:func:`numpy.true_divide` to `a` and
        `b`, except that elements corresponding to zeros in `b` are set
        to `zero` instead of actually being computed.
    """
    mask = (b != 0)
    result = numpy.true_divide(a, b, where=mask, out=out)
    if result.ndim > 0:
        result[~mask] = zero
    elif ~mask:
        result = zero
    return result


_thresholding_directions = {
    'le': numpy.less_equal,    '<=': numpy.less_equal,
    'lt': numpy.less,          '<': numpy.less,
    'ge': numpy.greater_equal, '>=': numpy.greater_equal,
    'gt': numpy.greater,       '>': numpy.greater,
}

_thresholding_types = {
    'std': lambda x, n: numpy.mean(x) + n * numpy.std(x),
    'iqr': lambda x, n: numpy.median(x) + n * iqr(x),
    'rms': lambda x, n: n * numpy.sqrt(numpy.square(x).mean()),
    'raw': lambda x, n: n,
}

def threshold(arr, thresh=3, type='std', direction='le'):
    """
    Apply a threshold to an array (usually an image).

    Parameters
    ----------
    arr : array-like
        The array to threshold.
    direction : str
        Which direction is considered passing:

        - ``'le'`` or ``'<='``: Elements of `arr` <= the threshold are
          marked `True`.
        - ``'lt'`` or ``'<'``: Elements of `arr` < the threshold are
          marked `True`.
        - ``'ge'`` or ``'>='``: Elements of `arr` >= the threshold are
          marked `True`.
        - ``'gt'`` or ``'>'``: Elements of `arr` > the threshold are
          marked `True`.

        The default is ``'le'``.
    thresh : array-like, optional
        The threshold value to apply. Must broadcast to the shape of the
        array. The exact meaning of the value is determined by `type`.
        The default is ``3`` (for 3-sigma thresholding).
    type : str, optional
        The type of threshold to use:

        - ``'std'``: Mean plus `threshold` times standard deviation.
        - ``'iqr'``: Median plus `threshold` times interqartile range.
        - ``'rms'``: `threshold` times the root-mean square.
        - ``None``, ``''``, ``'raw'``: Use `threshold` as-is.

        The default is 'std'.

    Returns
    -------
    numpy.ndarray :
        A boolean array of the same size and shape as `arr`, containing
        a mask indicating which elements pass threshold.
    """
    dmethod = option_lookup('direction', _thresholding_directions,
                            direction, key_func=str.lower)
    
    tmethod = option_lookup('type', _thresholding_types,
                            type or 'raw', key_func=str.lower)

    arr = numpy.asanyarray(arr)

    return dmethod(arr, tmethod(arr, thresh))


def ang_diff_pos(theta1, theta2, full=2.0 * numpy.pi):
    r"""
    Find the positive angular difference from `theta1` to `theta2`,
    normalized to :math:`[0, 2 \pi)`.

    The positive difference is the angle going in the positive
    direction from `theta1` to `theta2`, normalized to be in the range
    :math:`[0, 2 \pi)`.

    The return value can be computed without branching as ::

        ang_diff_pos = fmod(fmod(theta2 - theta1, full) + full, full)

    Inputs can be scalars or arrays. Arrays must broadcast together.

    Parameters
    ----------
    theta1 : array-like
        The start angle or angles, in radians.
    theta2 : array-like
        The end angle or angles, in radians.
    full : float
        The period of a full circle. Defaults to :math:`2 \pi`. Use 360
        for data in degrees, 400 for gradians, 6400 for mils, etc.

    Returns
    -------
    numpy.ndarray :
        An array containing the broadcasted positive normalized
        difference of the two inputs.
    """
    return numpy.fmod(numpy.fmod(theta2 - theta1, full) + full, full)


def ang_diff_min(theta1, theta2, full=2.0 * numpy.pi):
    r"""
    Find the angular difference from `theta1` to `theta2`, with the
    minimum absolute value normalized to :math:`[-\pi, \pi)`.

    The positive difference is the angle going in the positive
    direction from `theta1` to `theta2`, normalized to be in the range
    :math:`[0, 2 \pi)`. The negative difference is the angle going in
    the negative direction. This function returns the smaller of the
    two by absolute value.

    The return value can be computed without branching by rotating by
    half a circle before applying the moduli, then rotating back::

        ang_diff_min = fmod(fmod(theta2 - theta1 + 0.5 * full, full) +
                            full, full) - 0.5 * full

    Inputs can be scalars or arrays. Arrays must broadcast together.

    Parameters
    ----------
    theta1 : array-like
        The start angle or angles, in radians.
    theta2 : array-like
        The end angle or angles, in radians.
    full : float
        The period of a full circle. Defaults to :math:`2 \pi`. Use 360
        for data in degrees, 400 for gradians, 6400 for mils, etc.

    Returns
    -------
    numpy.ndarray :
        An array containing the broadcasted sign-preserving normalized
        difference of the two inputs with the smallest absolute value.
    """
    half = 0.5 * full
    # Broken out for readability
    step1 = numpy.fmod(theta2 - theta1 + half, full)
    return numpy.fmod(step1 + full, full) - half


def ang_diff_abs(theta1, theta2, full=2.0 * numpy.pi):
    r"""
    Find the absolute value of the minimum angular difference from
    `theta1` to `theta2`, normalized to :math:`[0, \pi)`.

    The minimum absolute difference is the smallest angle to get from
    `theta1` to `theta2` going in either direction, normalized to be in
    the range :math:`[0, \pi)`.

    The return value can be computed without branching as ::

        ang_diff_abs = abs(ang_diff_min(theta1, theta2, full))

    Inputs can be scalars or arrays. Arrays must broadcast together.

    Parameters
    ----------
    theta1 : array-like
        The start angle or angles, in radians.
    theta2 : array-like
        The end angle or angles, in radians.
    full : float
        The period of a full circle. Defaults to :math:`2 \pi`. Use 360
        for data in degrees, 400 for gradians, 6400 for mils, etc.

    Returns
    -------
    numpy.ndarray :
        An array containing the broadcasted minimum absolute normalized
        difference of the two inputs.
    """
    return numpy.abs(ang_diff_min(theta1, theta2, full))


def rms(arr, axis=None, bias=0, weights=None, ddof=0, out=None):
    """
    Comute the root-mean-square (RMS) of an array about an arbitrary
    bias.

    Parameters
    ----------
    arr : array-like
        The input array.
    axis : int or tuple or None, optional
        The axis or axes to compute the RMS along. The average of the
        squares will be taken along this axis or axes. None (the
        default) indicates the entire raveled array.
    bias : array-like, optional
        The offset about which to compute the RMS. The default is zero
        for standard RMS. The bias must broadcast to `arr`. Setting
        ``bias=numpy.mean(arr, axis=axis, keepdims=True)`` is equivalent
        to computing ``numpy.std(arr, axis=axis)``.
    weights : array-like, optional
        The weights to apply in the mean portion. Similar behavior to
        :py:func:`numpy.average`. Assumed to be all ones if omitted. If
        `axis` is None, must broadcast to `arr`. Otherwise, must be
        the same shape as either `arr[axis]` or `arr`.
    ddof : scalar, optional
        The degrees of freedom to subtract from the denominator. Defaults
        to zero.
    out : numpy.ndarray or None, optional
        The array to output results into, if not None. Use carefully,
        as *all* intermediate calculations will be placed into this
        array: a square, a mean and a square root. Using an output of
        an inappropriate type may lead to errors. The default is None.

    Return
    ------
    rms : numpy.ndarray
        The RMS of `arr` about `bias` along `axis`.
    """
    sq = numpy.square(numpy.asanyarray(arr) - bias, out=out)
    weights = numpy.array([1]) if weights is None \
                                            else numpy.asanyarray(weights)
    if axis is None:
        sq = sq.ravel()
        weights = weights.ravel()
    else:
        axis = numpy.core.multiarray.normalize_axis_index(axis, sq.ndim)
        if weights.ndim < sq.ndim:
            weights = numpy.reshape(weights,
                                    weights.shape + (1,) * (sq.ndim - axis - 1))
    weights = numpy.broadcast_to(weights, sq.shape)
    return numpy.sqrt(numpy.sum(sq * weights, axis=axis, out=out) / (weights.sum(axis=axis) - ddof), out=out)


def segment_distance(p, p1, p2, axis=None, return_t=False, segment=True):
    r"""
    Find the distance between an N-dimensional point and a line or line
    segment.

    The distance from a point to a line in N dimensions is the length
    of a normal dropped to the line. Using the fact that the dot
    product of orthogonal vectors is we can find the point
    :math:`\vec{p}_0` on the line that corresponds to this location.

    First, parametrize the points on the line through parameter
    :math:`t` as

    .. math::

       \vec{\ell} = \vec{p}_1 + t (\vec{p}_2 - \vec{p}_1)

    Then set up the equation with dot-products and solve for :math:`t`:

    .. math::

       (\vec{p} - \vec{p}_0) \cdot (\vec{p}_2 - \vec{p}_1) = 0

    .. math::

       (\vec{p} - \vec{p}_1 - t (\vec{p}_2 - \vec{p}_1)) \cdot (\vec{p}_2 - \vec{p}_1) = 0

    .. math::

       t (\vec{p}_2 - \vec{p}_1) \cdot (\vec{p}_2 - \vec{p}_1) = (\vec{p} - \vec{p}_1) \cdot (\vec{p}_2 - \vec{p}_1)

    .. math::

       t = \frac{(\vec{p} - \vec{p}_1) \cdot (\vec{p}_2 - \vec{p}_1)}{\left\lVert\vec{p}_2 - \vec{p}_1\right\rVert ^ 2}

    .. math::

       \vec{p}_0 = \vec{p}_1 + \frac{(\vec{p} - \vec{p}_1) \cdot (\vec{p}_2 - \vec{p}_1)}{\left\lVert\vec{p}_2 - \vec{p}_1\right\rVert ^ 2} (\vec{p}_2 - \vec{p}_1)

    The value of :math:`t` represents the location of :math:`\vec{p}_0`
    in relationship to :math:`\vec{p}_1` and :math:`\vec{p}_2`: values
    in the range :math:`[0, 1]` are on the line segment, negative
    values are on the side closer to :math:`\vec{p}_1`, and values
    greater than one are on the side of the line closer to
    :math:`\vec{p}_2`.

    The value of :math:`t` at the closest approach can be returned by
    setting ``return_t=True``. The value returned in this case applies
    to the entire line, even if ``segment == True`` and the closest
    point is one of the endpoints of the line segment.

    Parameters
    ----------
    p : array-like
        The target point. Must broadscast to `p1` and `p2`.
    p1 : array-like
        The start of the line segment. Must broadcast to the same shape
        as `p` and `p2`.
    p2 : array-like
        The end of the line segment. Must broadcast to the same shape
        as `p` and `p1`.
    axis : int or None
        The axis corresponding to the point vectors in the broadcasted
        arrays. If `None`, all point arrays are raveled.
    return_t: bool
        If `True`, return an additional value indicating the parameter
        :math:`t` at the distance of closest approach along the line.
        This will be the same regargless of `segment`.
    segment: bool
        If `True`, find the nearest point on the line segment bounded
        by `p1` and `p2` rather than the line passing between them.

    Returns
    -------
    dist : float or ~numpy.ndarray
        Distance from `p` to the line or line segment passing through
        `p1` and `p2`. The shape of the result is the broadcasted shape
        of the inputs, collapsed along `axis`.

        Scalar if `axis is None` or the inputs are all one-dimensional.
    t : float or ~numpy.ndarray
        An array of the same shape as `dist` containing the value of
        parameter :math:`t` for each line. The parameter is the
        location of the normal from `p` to the line passing through
        `p1` and `p2`, regardless if the distance is to the line
        segment or the line.

        Returned only if `return_t` is set.
    """
    p, p1, p2 = numpy.broadcast_arrays(p, p1, p2)
    seg = p2 - p1
    norm2_seg = (seg * seg).sum(axis=axis, keepdims=True)
    t = ((p - p1) * seg).sum(axis=axis, keepdims=True) / norm2_seg
    p0 = p1 + t * seg

    if segment:
        dist = numpy.empty_like(p0)
        mask1 = t < 0
        mask2 = t > 1
        numpy.subtract(p0, p, where=~(mask1 | mask2), out=dist)
        numpy.subtract(p1, p, where=mask1, out=dist)
        numpy.subtract(p2, p, where=mask2, out=dist)
    else:
        dist = p0 - p
    dist = numpy.square(dist, out=dist).sum(axis=axis, keepdims=True)
    numpy.sqrt(dist, out=dist)

    if axis is None or seg.ndim == 1:
        dist = dist.item()
        t = t.item()
    else:
        dist = dist.squeeze(axis)
        t = t.squeeze(axis)

    if return_t:
        return dist, t
    return dist



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 12 Jan 2022: Initial Coding

"""
Pytest plugin for processing the haggis-specific command-line options.
"""


def pytest_addoption(parser):
    """
    Add options to the default command line.

    The following options are added:

    `--plots`
        Draw plots of x-values, y-values and fit comparisons. This
        option checks if matplotlib is installed, and issues a warning
        if not.

    """
    parser.addoption("--plots", action="store_true", default=False,
                     help="Generate graphical plots of input data")



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 20 Aug 2023: Initial Coding

"""
Utilities for processing command line arguments.

This module works in conjunction with Python's standard library
:py:mod:`argparse` module.
"""


__all__ = ['SmartHelpFormatter']


from argparse import HelpFormatter


class SmartHelpFormatter(HelpFormatter):
    """
    A help formatter class that can keep newlines. Currently, the
    :py:mod:`argparse` functionality can only use the default
    arguments. To change the defaults, extend this class with
    different values of the class attributes.

    Note: This class extends some private functionality in the
    API. There is no guarantee that it will work in newer versions of
    python. Please submit an issue if you experience any problems.

    .. py:attribute:: KEEP_NEWLINES

       If True, manually inserted newlines will be preserved along with
       standard line-wrapping behavior.

    .. py:attribute:: KEEP_LEADING

       If True, manually inserted leading spaces will be preserved
       while performing standard line-wrapping behavior and
       indentation.

    .. py:attribute:: TABWIDTH

       If :py:attr:`KEEP_LEADING` is True, this attribute determines
       the number of spaces by which a ``\\t`` character is replaced
       by, while preserving indentation.
    """
    KEEP_NEWLINES = True
    KEEP_LEADING = True
    TABWIDTH = 4

    def __init__(self, *args, keep_newlines=None, keep_leading=None,
                 tabwidth=None, **kwargs):
        super().__init__(*args, **kwargs)
        def check(value, default):
            return default if value is None else value
        self._keep_newlines = check(keep_newlines, self.KEEP_NEWLINES)
        self._keep_leading = check(keep_leading, self.KEEP_LEADING)
        self._tabwidth = check(tabwidth, self.TABWIDTH)

    def _get_indent(self, text):
        """
        Compute the number of leading spaces in a string. If
        ``self._keep_leading`` is False, return zero.
        """
        if self._keep_leading:
            n = len(text) - len(text.lstrip())
            n += (self._tabwidth - 1) * text[:n].count('\t')
        else:
            n = 0
        return n

    def _split_lines(self, text, width):
        """
        Overrides the behavior of the base class with the configured
        behavior for leading whitespace and newlines.
        """
        lines = text.splitlines() if self._keep_newlines else [text]
        r = []
        for line in lines:
            n = self._get_indent(line)
            r.extend(' ' * n + s
                     for s in super()._split_lines(line, width - n))
        return r

    def _fill_text(self, text, width, indent):
        """
        Overrides the behavior of the base class with the configured
        behavior for leading whitespace and newlines.
        """
        lines = text.splitlines() if self._keep_newlines else [text]
        r = []
        for line in lines:
            n = self._get_indent(line)
            r.append(super()._fill_text(line, width, indent + n * ' '))
        return '\n'.join(r)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 03 Nov 2022: Initial Coding

"""
Tests for the :py:mod:`haggis.npy_util` module.
"""


from pytest import raises

from ..recipes import chained_getter


class Namespace:
    """
    Child of object with arbitrarily assignable attributes and a
    convenience constructur.
    """
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TestChainedGetter:
    def test_basic(self):
        spec = 'a.b[1][].c.d[key][i].e[key].f[]'
        obj = Namespace(
            a=Namespace(
                b=[None, [1, 2, Namespace(
                    c=Namespace(
                        d={'key': [Namespace(
                            e={'key': Namespace(f={'inner': 'target'})}
                        )]}
                    )
                )]]
            )
        )
        getter = chained_getter(spec, 2, 'inner', key='key', i=0)

        assert getter(obj) == 'target'

    def test_circref(self):
        obj = Namespace(a=[], b={}, z='target')
        obj.c = obj
        obj.a.append(obj)
        obj.a.append(obj.a)
        obj.a.append(obj.b)
        obj.b['obj'] = obj
        obj.b['list'] = obj.a
        obj.b['dict'] = obj.b

        assert chained_getter('c.c.c.z')(obj) == 'target'
        assert chained_getter('a[0].c.a[1][1][0].z')(obj) == 'target'
        assert chained_getter('b[].c.b[key][key][][0].z', 'obj', 'list',
                              key='dict')(obj) == 'target'

    def test_missing_arg(self):
        raises(IndexError, chained_getter, '[][]', 'a')

    def test_extra_arg(self):
        obj = {'a': {'b': 'target'}}
        assert chained_getter('[][]', *'abc')(obj) == 'target'

    def test_missing_kwarg(self):
        raises(KeyError, chained_getter, '[i][j]', i='a')

    def test_extra_kwarg(self):
        obj = {'a': {'b': 'target'}}
        assert chained_getter('[i][j]', i='a', j='b', k='c')(obj) == 'target'

    def test_extra_args(self):
        obj = {'a': {'b': 'target'}}
        assert chained_getter('[i][]', *'bc', i='a', j='b')(obj) == 'target'

    def test_bad_index(self):
        # Missing list index
        raises(IndexError, chained_getter('[]', 0), [])
        # Non-numerical list index
        raises(TypeError, chained_getter('[]', 'a'), [])
        # Missing dictionary key
        raises(KeyError, chained_getter('[]', 'a'), {})
        # Unhashable dictionary key
        raises(TypeError, chained_getter('[]', []), {})

    def test_bad_attribute(self):
        raises(AttributeError, chained_getter('a.d.c'),
               Namespace(a=Namespace(b=Namespace(c='target'))))

    def invalid_identifier(self):
        # Unterminated bracket
        raises(SyntaxError, chained_getter, 'a[')
        # Unopended bracket
        raises(SyntaxError, chained_getter, 'a]')
        # Other junk
        raises(SyntaxError, chained_getter, 'a$#%')
        # Starts with number
        raises(SyntaxError, chained_getter, '0asdf')

    def test_dots(self):
        raises(AttributeError, chained_getter('.a'), Namespace(a='target'))
        raises(AttributeError, chained_getter('.[0]'), Namespace(a=['target']))
        raises(SyntaxError, chained_getter, '[0]a')
        raises(AttributeError, chained_getter('[0].[0]'), [Namespace(a=[])])



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


""" Configuration file for the Sphinx documentation builder. """

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('../src'))


# -- Project information -----------------------------------------------------

project = 'Haggis'
copyright = '2023, Joseph R. Fox-Rabinovitz'
author = 'Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>'

from haggis.version import __version__
# The short X.Y version
version = '.'.join(__version__.split('.')[:2])
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# Minimal Sphinx version requirement.
#
needs_sphinx = '1.7.1'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.imgmath',
    'sphinx.ext.napoleon',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', '_src', '_intersphinx', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
try:
    import sphinx_rtd_theme
except ImportError:
    html_theme = 'alabaster'
else:
    html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'HaggisDoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'Haggis.tex', 'Haggis Documentation',
     'Joseph R. Fox-Rabinovitz', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'haggis', 'Haggis Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Haggis', 'Haggis Documentation',
     author, 'Haggis', 'MS Word Document Generator',
     'Miscellaneous'),
]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Intersphinx configuration: attempt to download object listings from the
# official documentations, use a (possibly outdated) local backup if that fails.
intersphinx_mapping = {
    'astropy': (
        'https://docs.astropy.org/en/stable/',
        (None, '_intersphinx/astropy.inv')
    ), 'docx': (
        'https://python-docx.readthedocs.io/en/latest/',
        (None, '_intersphinx/docx.inv')
    ), 'matplotlib': (
        'https://matplotlib.org/',
        (None, '_intersphinx/matplotlib.inv')
    ), 'openpyxl': (
        'https://openpyxl.readthedocs.io/en/stable/',
        (None, '_intersphinx/openpyxl.inv')
    ), 'pandas': (
        'https://pandas.pydata.org/pandas-docs/stable/',
        (None, '_intersphinx/pandas.inv')
    ), 'pyqt6': (
        'https://www.riverbankcomputing.com/static/Docs/PyQt6/',
        (None, '_intersphinx/pyqt6.inv')
    ), 'python': (
        'https://docs.python.org/3/',
        (None, '_intersphinx/python.inv')
    ),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# Suppress trailing parentheses
add_function_parentheses = False



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for implementing, raising and handling exceptions.
"""

__all__ = ['ErrorTransform']


class ErrorTransform:
    """
    A context manager that translates any exceptions into a different
    type with a predefined message. The constructor arguments are
    assigned to class attributes directly.

    .. py:attribute:: in_type

       The exceptions to look for, in any format accepted by an
       ``except`` clause: a single type or a tuple of types.

    .. py:attribute:: out_type

       The type to reraise as.

    .. py:attribute:: message

       A format string containing the message of the rethrown error. The
       string is expected to conform to the :ref:`formatspec`.

    .. py:attribute:: args

       Additional positional arguments to pass to
       :py:meth:`message.format <str.format>`.

    .. py:attribute:: kwargs

       Additional keyword arguments to pass to `message.format`. There
       three dynamic keywords are always passed in:

         - ``type``: The class of the trapped error.
         - ``str``: The result of ``str(exc)`` on the trapped error.
         - ``repr``: The result of ``repr(exc)`` on the trapped error.

       These three names must not appear as keys in `kwargs`.
    """
    def __init__(self, in_type, out_type, message, *args, **kwargs):
        """
        Construct a context manager that will transform `in_type`
        exceptions to `out_type` with the specified formatted message.
        """
        self.in_type = in_type
        self.out_type = out_type
        self.message = message
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def KeyError(cls, out_type, name='dictionary'):
        """
        Create an instance of this class suitable for handling
        occurrences of :py:exc:`KeyError` caused by, e.g. dictionary
        access via :py:meth:`~object.__getitem__`.

        :py:attr:`in_type` is implicitly :py:exc:`KeyError` and `name`
        is the name of the dictionary where the error occurred.
        """
        return cls(KeyError, out_type, "Missing '{str}' from {name}",
                   name=name)

    def __enter__(self):
        """
        Enter the context manager (no-op).
        """
        pass

    def __exit__(self, type, value, traceback):
        """
        Exit the context manager.

        If an exception of type :py:attr:`in_type` occurred, it is
        transformed to an exception of type :py:attr:`out_type` with the
        configured message and raised. All other exception types are
        passed through (this method returns :py:obj:`None`).
        """
        if type is not None and issubclass(type, self.in_type):
            raise self.out_type(
                self.message.format(*self.args, **self.kwargs, type=type,
                                    str=str(value), repr=repr(value))
            ) from value



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Utilities for working with zip archives.
"""

__all__ = ['filter', 'remove']


from itertools import repeat
from os import remove as delete
from os.path import splitext
from shutil import copyfileobj
from tempfile import mkstemp
from warnings import warn
from zipfile import ZipFile, ZipInfo


def remove(zipname, *filenames):
    """
    Remove the specified file from the named zip archive.

    Elements of `filenames` may be strings or
    :py:class:`zipfile.ZipInfo` objects. In the latter case, only the
    :py:attr:`~zipfile.ZipInfo.filename` attribute is used to identify
    matches. The other metadata is ignored.

    Raise a warning for any file names that are not found. Actual
    removal is done by recreating the archive minus the filtered
    elements in a temporary file, then overwriting the original with it.
    """
    return filter(zipname, *filenames, filter=None)


def filter(zipname, *filenames, filter=None):
    """
    .. py:function:: filter(zipname, *filenames, filter=None)
       :noindex:
    .. py:function:: filter(zipname, filterDict)
       :noindex:

    Modify the contents of a file or files in the specified zip archive.

    A filter value of :py:obj:`None` removes the selected files.

    There are two calling conventions for this function. In the first
    case, it accepts a sequence of file names or
    :py:class:`zipfile.ZipInfo` objects. `filter` is a function that
    accepts a byte string with the decompressed file contents and
    returns the filtered string to replace the contents with. The
    filtered string may be a true string or bytes.

    In the second case, `filter` is not provided separately, but rather
    as the values in a mapping. The keys are the file names to filter.
    This version is only activated when there is a single additional
    argument besides `zipname`. In this case `filter` is completely
    ignored.

    File contents ares re-inserted with the same metadata as the
    original.
    """
    if len(filenames) != 1 or isinstance(filenames[0], str) or \
                isinstance(filenames[0], ZipInfo):
        iter = zip(filenames, repeat(filter))
    else:
        iter = filenames[0].items()

    filenames = {
        (name if isinstance(name, str) else name.filename): {
            'count': 0, 'filter': filter
        } for name, filter in iter
    }

    # Based heavily on http://stackoverflow.com/a/4653863/2988730 and
    # http://stackoverflow.com/a/25739108/2988730.
    temp_file, temp_name = mkstemp(suffix=splitext(zipname)[1])
    with ZipFile(zipname, 'a') as zin, \
                ZipFile(open(temp_file, 'wb'), 'w') as zout:
        zout.comment = zin.comment
        for item in zin.infolist():
            if item.filename in filenames:
                fileItem = filenames[item.filename]
                fileItem['count'] += 1
                if fileItem['filter'] is None:
                    # Delete file
                    data = None
                else:
                    # Filter file
                    data = fileItem['filter'](zin.read(item))
            else:
                # Passthru file
                data = zin.read(item)

            if data is not None:
                # Write back filtered data. Don't skip empty strings (only None).
                zout.writestr(item, data)

    # Doing it this way will preserve the file permissions
    with open(zipname, 'wb') as fout, open(temp_name, 'rb') as fin:
        copyfileobj(fin, fout)
    delete(temp_name)

    for name, item in filenames.items():
        count = item['count']
        if count > 1:
            warn('"{}" appeared "{}" times '
                 'in "{}"'.format(name, count, zipname))
        elif count == 0:
            warn('"{}" not found in "{}"'.format(name, zipname))




# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2024  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 29 Jun 2024: Initial Coding

"""
Utilities for facilitating common operations with different GUI packages.

The module names in this package generally correspond to the informal
name of the GUI package in question.
"""



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Functions for manipulating the structure of objects.

This includes things like spoofing module contents, copying classes and
functions, and automatically creating properties.
"""

__all__ = [
    'HiddenPropMeta', 'copy_func', 'copy_class', 'getsizeof',
    'package_root', 'size_type_mapping', 'update_module',
]


from array import array
from collections.abc import Iterable, Mapping
from copy import copy
from functools import update_wrapper, WRAPPER_ASSIGNMENTS
from itertools import chain
from os.path import basename, dirname
import sys
from types import FunctionType, ModuleType


#: List mapping of types to the special processing routines required to
#: support them with :py:func:`getsizeof`.
#:
#: Types are checked from the end of the list. The first element is
#: :py:class:`~collections.abc.Iterable`, which is the universal
#: catchall. Later elements are more specific types. Custom types
#: should be appended to the end.
#:
#: The following types are supported out of the box:
#:
#: - :py:class:`~collections.abc.Iterable`
#: - :py:class:`~collections.abc.Mapping`
#: - :py:class:`str`, :py:class:`bytes`, :py:class:`bytearray`,
#:   :py:class:`array.array`
#: - :py:class:`numpy.ndarray`
#:
#: The list contains two-element tuples, as would be used to initialize
#: a :py:class:`dict`. The first element can be a scalar type or tuple
#: of types. The second element may be `None`, indicaing a passthrough
#: to :py:func:`sys.getsizeof`, or a callable accepting an object of
#: the correct type, returning an iterable of elements in any order.
#: The callable only needs to iterate the top-level elements.
#:
#: Permanently register handlers by appending the appropriate tuple to
#: this list. Temporarily register them by using the `handlers`
#: argument to :py:func:`getsizeof`.
size_type_mapping = [
    (Iterable, iter),
    (Mapping, lambda m: (e for i in m.items() for e in i)),
    ((str, bytes, bytearray, array), None),
]


try:
    import numpy
except ImportError:
    pass
else:
    def ndarray_handler(x):
        """
        Yield all the elements of the array that are objects.

        Structured dtypes are parsed field-by-fiels, with proper
        handling of nested structures.

        Parameter
        ---------
        x : numpy.ndarray
            The array to parse.

        Return
        ------
        elem :
            A generator of all the elements of `x` that have dtype code
            ``'O'``.
        """
        def dfs(x):
            """
            DFS chosen because nested datatypes are expected to be of
            limited deoth, and because it's most likely to return the
            fields and subarrays in order of increasing offset.

            Parameter
            ---------
            x : numpy.ndarray
                An array potentially containing objects.

            Return
            ------
            elem :
                An generator chaining the elements of the array that
                are objects together.
            """
            dtype = x.dtype
            if not dtype.hasobject:
                return

            # Structure
            if dtype.fields is not None:
                for name in dtype.fields:
                    yield from dfs(x[name])
            # Array
            elif dtype.subdtype is not None:
                for i in numpy.arange(numpy.prod(dtype.shape)):
                    yield from dfs(x[i])
            # Primitive
            else:
                yield from map(numpy.ndarray.item, numpy.nditer(x, flags=['refs_ok']))

        yield from dfs(x)

    size_type_mapping.append((numpy.ndarray, ndarray_handler))


def getsizeof(obj, handlers=None, default=sys.getsizeof(int)):
    """
    Recursive version of :py:func:`sys.getsizeof` for handling
    iterables and mappings.

    Supports automatic circular reference detection, and does not
    double-count repeated references. String and array types get
    special treatement: they are iterable, but not processed
    recursively because their size already includes the buffer. The
    following types are treated as array types:

    - :py:class:`str`
    - :py:class:`bytes`
    - :py:class:`bytearray`
    - :py:class:`array.array`
    - :py:class:`numpy.ndarray`

    Additional array/string-like types may be added by appending them
    to the module-level tuple :py:attr:`size_type_mapping`. Numpy
    arrays require special treatment because they can contain
    references to other objects nested at arbitrarily deep levels of
    the datatype.

    References are not fully supported yet, but a custom handler can
    be added to :py:attr:`size_type_mapping`. Object attributes have
    only rudimentary support via recursion into ``__dict__`` and
    ``__slots__`` (not necessarily mutually exclusive). Additional
    support is available via custom implementations of
    :py:meth:`~object.__sizeof__`, or through custom handlers.

    Parameters
    ----------
    obj :
        The object whose size is to be computed.
    handlers : None, Iterable[tuple[type, callable]], Mapping
        Mapping of types to handler functions, or list of tuples
        containing type-handler pairs. Items are iterated in reverse
        order, so place more specific types last. Callables must
        accept the object whose elements are to be sized, and return
        an iterable of the top-level elements. Any handlers speciied
        through this argument supersede defaults set in
        :py:attr:`size_type_mapping`.
    default : int
        The default size to use for objects that do not support a
        :py:meth:`~object.__sizeof__` operation. Default:
        ``sys.getsizeof(int)``.

    Return
    ------
    size : int
        The size of the object and all the references it contains.
        This is especially useful for container types.

    Notes
    -----
    This recipe is inspired by Raymond Hettinger's "Compute Memory
    footprint of an object and its contents" available at
    https://github.com/ActiveState/recipe-577504-compute-mem-footprint
    and https://code.activestate.com/recipes/577504/. This function was
    originally written at https://stackoverflow.com/a/70793151/2988730.
    Things I took from Raymond's recipe after the fact:

    - Making handlers iterate through the elements instead of applying
      the original recursion function directly.
    - Using a default value.
    - Accepting a mapping of extra handlers.

    Things I added:

    - Proper handling of strings, bytes and bytearrays
    - Numpy array handler
    - Global type registry
    - Support for `__dict__` and `__slots__`
    """
    if handlers is None:
        handlers = ()
    elif isinstance(handlers, Mapping):
        handlers = handlers.items()

    seen = set()
    def recurse(obj):
        x = id(obj)
        if x in seen:
            return 0
        seen.add(x)
        size = sys.getsizeof(obj, default)

        for type, method in chain(reversed(handlers),
                                  reversed(size_type_mapping)):
            if isinstance(obj, type):
                if method is not None:
                    for elem in method(obj):
                        size += recurse(elem)
                break

        if hasattr(obj, '__dict__'):
            size += recurse(vars(obj))
        for slot in getattr(obj, '__slots__', ()):
            if hasattr(obj, slot):
                size += recurse(getattr(obj, slot))
        return size

    return recurse(obj)


def package_root(module):
    """
    Find the directory containing the root package in which a module is
    defined.

    Only works for modules with a valid ``__file__`` attribute.

    Parameters
    ----------
    module : str or ~types.ModuleType
        The module to investigate. If a name is passed in, the module
        must exist in :py:obj:`sys.modules`.

    Return
    ------
    path : str
        The root path of the package containing the module.
    """
    if isinstance(module, str):
        module = sys.modules[module]
    name = module.__name__
    path = module.__file__
    if basename(path) == '__init__.py':
        path = dirname(path)
    for _ in range(name.count('.') + 1):
        path = dirname(path)
    return path


def update_module(current, other, recurse=False):
    """
    Updates the dict of the module `current` with the dict of
    `other`.

    Either input may be a string (full name as given by ``__name__``),
    or a module reference.

    All functions and classes in `other` whose module is `other`
    will by default be copied and reassigned to `current`. All other
    non-dunder attributes will be copied exactly. Dunder attributes will
    be skipped, except ``__all__``, which will be shallow-copied as a
    list.

    Any attributes already defined in `current` will be skipped. This
    ensures that the globals defined in new methods will be updated
    correctly for the new module.

    If `recurse` is set to :py:obj:`True`, any sub-modules of `other`
    will be copied using this method instead of referenced directly.
    """
    def getmod(mod):
        if isinstance(mod, str):
            return sys.modules[mod]
        # Ducktype non-string objects. This allows things like
        # classes spoofing a module to pass through just fine.
        return mod
    current = getmod(current)
    other = getmod(other)
    cn, on = current.__name__, other.__name__
    cd = current.__dict__
    for name, item in other.__dict__.items():
        if name in cd:
            continue
        if name.startswith('__') and name.endswith('__'):
            if name == '__all__':
                try:
                    item = item[:]
                except Exception:
                    pass
            else:
                continue
        elif isinstance(item, FunctionType):
            if item.__module__ == on:
                item = copy_func(item, cd, cn)
        elif isinstance(item, type):
            if item.__module__ == on:
                item = copy_class(item, cd, cn)
        elif isinstance(item, ModuleType) and recurse:
            if item.__name__.startswith(on + '.'):
                mod = ModuleType(item.__name__.replace(on, cn), item.__doc__)
                update_module(mod, item, recurse)
                sys.modules[mod.__name__] = mod
                item = mod
        cd[name] = item


def copy_class(c, globals=None, module=None):            
    """
    Creates a shallow copy of a class object, optionally replacing
    its module and the object its methods reference for globals.

    This function is useful when importing a class into another
    module, and having it behave as a class of the importing module::

        from mod import cls
        cls = copy_class(cls, globals(), __name__)

    Parameters
    ----------
    c : type
        The class to copy
    globals : dict or None
        If :py:obj:`None`, copy the global dictionaries referenced by
        the methods unchanged. A popular alternative is ``globals()``.
    module : str or None
        The name of the module that this class belongs to. If
        :py:obj:`None`, keep ``c.__module__`` and the modules of all
        methods directly. A popular alternative is ``__name__``.

    Notes
    -----
    This function may not work properly for classes whose metaclass does
    not invoke :py:meth:`type.new <object.__new__>` at some point in the
    construction process.

    Based on https://stackoverflow.com/a/49157516/2988730.
    """
    if module is None:
        module = c.__module__
    attrs = dict(c.__dict__)
    for name, item in c.__dict__.items():
        if isinstance(item, type):
            attrs[name] = copy_class(item, globals, module)
        elif isinstance(item, FunctionType):
            attrs[name] = copy_func(
                item, item.__globals__ if globals is None else globals, module
            )
    d = type.__new__(type(c), c.__name__, c.__bases__, attrs)
    d.__module__ = module
    return d


def copy_func(f, globals=None, name=None, module=None):
    """
    Creates a shallow copy of a function object, optionally replacing
    the object it references for its globals.

    This function is useful when importing a function into another
    module, and having it behave as a function of the importing module::

        from mod import func
        func = copy_func(func, globals(), module=__name__)

    Parameters
    ----------
    f : function
        The object to copy
    globals : dict or None
        If :py:obj:`None`, copy the global dictionary referenced by
        `f`. A popular alternative is ``globals()``.
    name : str or None
        The name to assign to the new function. If None, copy
        ``f.__name__`` directly.
    module : str or None
        The name of the module that this function belongs to. If
        :py:obj:`None`, copy ``f.__module__`` directly. A popular
        alternative is ``__name__``.

    Notes
    -----
    Based originally on https://stackoverflow.com/a/13503277/2988730,
    and updated in https://stackoverflow.com/a/49077211/2988730.
    """
    g = FunctionType(f.__code__,
                     f.__globals__ if globals is None else globals,
                     name=f.__name__ if name is None else name,
                     argdefs=f.__defaults__, closure=f.__closure__)
    g = update_wrapper(g, f)
    if module is not None:
        g.__module__ = module
    g.__kwdefaults__ = copy(f.__kwdefaults__)
    return g


class HiddenPropMeta(type):
    """
    Creates a class with "hidden" read-only properties named in the
    :py:attr:`__hidden_properties__` attribute.

    A hidden property is one that stores its value under a 
    :py:attr:`~object.__dict__` key with the same name. This meta-class
    is therefore incompatible with anything that uses
    :py:obj:`~object.__slots__`.

    .. py:attribute:: __hidden_properties__

       This can be a single string, an iterable of strings, or an
       iterable of two-element tuples containing a string name and an
       initial value. Strings and tuples may be mixed together in an
       iterable. The attribute will be removed from the class body by
       this metaclass after it is processed.

    If the class has an explicit ``__init__`` method defined, it will be
    properly decorated to set the default values of the hidden
    properties. If an explicit ``__init__`` is not found, the implicit
    ``super.__init__`` constructor will be decorated in the same way and
    set as the initializer.
    """
    def __new__(meta, name, bases, body, **kwargs):
        if '__hidden_properties__' in body:
            props = body['__hidden_properties__']
            del body['__hidden_properties__']
            if isinstance(props, str):
                props = [props]

            def check_prop(prop):
                """
                Check if `prop` is a string or a 2-element tuple and
                returns a 2-element tuple. The first element must be a
                string, the second defaults to :py:obj:`None`.
                """
                if isinstance(prop, str):
                    return (prop, None)
                if not isinstance(prop, tuple) or len(prop) != 2 or \
                   not isinstance(prop[0], str):
                    raise ValueError(
                        'property description must be name or name-value tuple'
                    )
                return prop

            props = dict(check_prop(prop) for prop in props)

            def fget(name):
                """
                Create the getter for a given property name.
                """
                def getter(self):
                    return self.__dict__[name]
                getter.__name__  = name
                getter.__doc__ = f'Retrieves {name} as a read-only property.'
                return getter

            for prop in props:
                if prop not in body:
                    body[prop] = property(fget(prop))
                elif not isinstance(body[prop], property):
                    raise ValueError(
                        'property {!r} shadows non-property '
                        'attribute'.format(prop)
                    )
        else:
            props = None

        cls = super().__new__(meta, name, bases, body)
        if props:
            def wrapper(parent, initializer):
                def  __init__(self, *args, **kwargs):
                    self.__dict__.update(initializer)
                    return parent(self, *args, **kwargs)

                update_wrapper(
                    __init__, parent, tuple(
                        x for x in WRAPPER_ASSIGNMENTS if x != '__qualname__'
                    )
                )
                __init__.__qualname__ = '.'.join(
                    (cls.__qualname__, '__init__')
                )
                return __init__
            setattr(cls, '__init__', wrapper(getattr(cls, '__init__'), props))
        return cls



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 12 Jan 2022: Initial Coding

"""
Collection of testing utilities for common actions.
"""


from contextlib import contextmanager
from os.path import join


FOLDER = '.haggis_test'


@contextmanager
def plotting_context(*args, **kwargs):
    """
    Context manager that imports matplotlib as necessary, creates a
    figure, and closes it on exit.

    All parameters are passed directly to
    :py:func:`~matplotlib.pyplot.figure`.
    """
    from matplotlib import pyplot as plt
    figure = plt.figure(*args, **kwargs)
    yield figure
    plt.close(figure)


def save(fig, module, label, debug=False, dpi=300):
    """
    Saves a figure in the standard test output directory.

    The figure name is nomalized with the name of the calling module.
    """
    ext = 'debug.png' if debug else 'png'
    label = f'{module}-{label}.{ext}'

    fig.savefig(join(FOLDER, label), dpi=dpi)



# -*- coding: utf-8 -*-

# haggis: a library of general purpose utilities
#
# Copyright (C) 2023  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding

"""
Useful and mostly compact shortcuts for common operations.

Implementations based on some of the recipes provided in the Python
documentation, and other sources like Stack Overflow.
"""

__all__ = [
    'all_combinations', 'all_nsc', 'any_nsc', 'chained_getter',
    'CloseableMixin', 'consume', 'grouper', 'immutable', 'islast',
    'is_ordered_subset', 'KeyedSingleton', 'lenumerate', 'RangeBuilder',
    'shift_up',
]


from collections import deque
from functools import reduce
from itertools import chain, combinations, islice, repeat
from operator import index, or_, and_, attrgetter, itemgetter
import re

try:
    from contextlib import AbstractContextManager
except ImportError:
    class AbstractContextManager:
        def __enter__(self):
            return self

        def __exit__(self, *args):
            return None


class CloseableMixin(AbstractContextManager):
    """
    Mixin for simple context management for objects with a
    :py:meth:`close` method.
    """

    def __exit__(self, *args):
        """
        Closes the object, returns `None`.
        """
        self.close()

    def close(self):
        """
        Default no-op implementation of `close`.

        Child classes are always expected to provide the correct
        behavior.
        """
        pass


def immutable(allow_properties=True,
              docstring='Forbids write access to class attributes.'):
    r"""
    Create a class decorator that sets a pre-defined version of
    :py:meth:`~object.__setattr__` that forbids write access to the
    attributes of a class.

    Access to properties may be allowed through `allow_properties`.
    The default setting of :py:obj:`True` does not guarantee that
    properties will be acessible, just that the parent implementation
    of :py:meth:`~object.__setattr__` will be responsible for handling
    the access.

    The messages of exceptions that are raised are based on the type of
    access that is requested. They mimic the default message that appear
    from most Python classes.

    Parameters
    ----------
    allow_properties : bool
        Whether or not to make an exception for properties. If
        :py:obj:`True`, setting of properties will be delegated to the
        parent's :py:meth:`~object.__setattr__`.
    docstring : str
        The string to associate with :py:meth:`~object.__setattr__`\ 's
        `__doc__` attibute.

    Return
    ------
    decorator : callable
        A class decorator that inserts a :py:meth:`~object.__setattr__`
        function into the decorated class.

    Notes
    -----
    Checking for properties based on the following Stack Overflow post:
    https://stackoverflow.com/a/46101204/2988730. See also the
    discussion following the answer itself.

    MRO issues with properly calling :py:meth:`~object.__setattr__` on
    properties are discussed in great detail in the answers and
    discussion surrounding https://stackoverflow.com/q/46121637/2988730.
    """
    def make_setattr(owner):
        def __setattr__(self, name, value):
            if hasattr(self, name):
                msg = 'readonly attribute'
            else:
                msg = '{!r} object has no attribute {!r}'.format(
                    type(self).__name__, name
                )
            raise AttributeError(msg)

        if allow_properties:
            default_setattr = __setattr__

            def __setattr__(self, name, value):
                typ = type(self)
                prop = getattr(typ, name, None)
                if isinstance(prop, property):
                    if owner is None:
                        return prop.__set__(self, value)
                    return super(owner, self).__setattr__(name, value)

                default_setattr(self, name, value)

        if docstring:
            __setattr__.__doc__ = docstring

        return __setattr__

    def decorator(cls):
        """
        A class decorator that assigns a special
        :py:meth:`~object.__setattr__` method to forbid attribute
        modification.
        """
        cls.__setattr__ = make_setattr(cls)
        return cls

    return decorator


def consume(iterator, n=None):
    """
    Consume an iterator entirely, or advance `n` steps ahead.

    This function is based pretty much exactly on the similarly named
    recipe in the documentation at
    https://docs.python.org/3/library/itertools.html#itertools-recipes
    """
    # Use functions that consume iterators at C speed.
    if n is None:
        # feed the entire iterator into a zero-length deque
        deque(iterator, maxlen=0)
    else:
        # advance to the empty slice starting at position n
        next(islice(iterator, n, n), None)


def grouper(iterable, n, tail=True, fillvalue=None):
    """
    Gather an iterable into chunks or blocks of fixed length.

    Based on ``grouper`` from :ref:`itertools-recipes`. The main
    difference is that this implementation allows the final chunk to
    contain fewer elements. This version also returns lazy iterables for
    the inner chunks, unlike the ``grouper`` recipe which returns
    chunks.

    By default, the iterator assumes that all of the elements of a group
    have been consumed by the caller. If this is not the case, the next
    group will start with the un-consumed segment. An easy way to
    guarantee proper consumption is to use :py:func:`consume`.

    Inspired by https://stackoverflow.com/a/23926929/2988730.

    Parameters
    ----------
    iterable : iterable
        The iterable to split.
    n : int
        The size of the chunks to split into.
    tail : bool
        If :py:obj:`True`, the final group may contain fewer than `n`
        elements. If :py:obj:`False`, `fillvalue` will be used to pad
        the last group if it turns out to be shorter than `n`.
    fillvalue :
        The item to pad the final chunk with if `tail` is
        :py:obj:`False` and the iterable's length is not a multiple of
        `n`.

    Yields
    ------
    group :
        An iterable containing the next `n` or fewer elements of
        `iterable`.

    Notes
    -----
    This uses the updated generator protocol as per
    https://stackoverflow.com/a/45605358/2988730, meaning that this
    function theoretically requires Python 3.5+.
    """
    iterable = iter(iterable)
    n = index(n)
    while True:
        try:
            peek = next(iterable)
        except StopIteration:
            return
        it = iterable if tail else chain(iterable, repeat(fillvalue))
        yield chain((peek,), islice(it, n - 1))


def islast(iterable):
    """
    An iterator similar to the builtin :py:func:`enumerate`, except that
    instead of the index, it returns a boolean flag indicating if the
    current element is the last one.

    Based on the recipe at http://stackoverflow.com/a/2429118/2988730.
    Saw http://stackoverflow.com/a/2429260/2988730 after writing this.
    """
    iterable = iter(iterable)
    try:
        item = next(iterable)
    except StopIteration:
        return None
    for peek in iterable:
        yield False, item
        item = peek
    yield True, item


def lenumerate(iterable, seq_type=None):
    """
    A generator that returns the length of each element of `iterable` in
    addition to the element itself.

    Each element must be an iterable itself. `seq_type` controls the
    type of sequence the elements will be turned into to get the length.
    If `None`, elements are assumed to have a `len` and will not be
    altered. Otherwise, `seq_type` must be a callable that accepts one
    iterable argument and returns a valid sequence, like
    :py:class:`list`, :py:class:`tuple`, :py:class:`str` or
    :py:class:`bytearray`.

    Specifying `seq_type` is roughly equivalent to calling
    ``lenumerate(map(seq_type, iterable))``. It guarantees consumption
    of all the elements of the sub-iterables. This could be important
    for something like :py:func:`grouper`.

    The name is a portmanteau of :py:func:`len` and
    :py:func:`enumerate`.
    """
    if seq_type is None:
        seq_type = lambda x: x
    for item in iterable:
        item = seq_type(item)
        yield len(item), item


def is_ordered_subset(a, b):
    """
    Check if iterable `a` is an ordered subset of iterable `b`.

    An ordered subset means that elements can be selected from `b`
    while preserving order in such a way as to get `a`.

    For example,::

        0, 1, 2

    is an ordered subset of::

        0, 2, 1, 3, 2

    but::

        0, 3, 1

    is not.

    Source: https://stackoverflow.com/a/11820887/2988730
    """
    it = iter(b)
    return all(x in it for x in a)


def all_combinations(sequence, start=0, stop=None):
    """
    Generate all combinations of a sequence with all possible lengths.

    The smallest length is `start`, which defaults to zero. The largest
    is `stop`, which defaults to `len(sequence)`. `stop` is inclusive.

    To use an iterable with no :py:func:`len`, specify `stop` manually.

    Example::

        all_combinations([1, 2, 3])

    generates::

        (,), (1,), (2,), (3,), (1, 2), (1, 3), (2, 3), (1, 2, 3)
    """
    if stop is None:
        stop = len(sequence)
    for n in range(start, stop + 1):
        for combo in combinations(sequence, n):
            yield combo


class KeyedSingleton(type):
    """
    A meta-class for genralized singleton initialization that returns an
    existing object based on an input key.

    The key is expected to be the first argument to the class
    constructor.

    Inspired by https://stackoverflow.com/a/8665179/2988730 and
    https://stackoverflow.com/a/31723324/2988730.

    I added my own, improved answer at
    https://stackoverflow.com/a/45175660/2988730
    """
    def __init__(cls, *args, **kwargs):
        """
        Initializes the class with an empty dictionary of instances.

        Arguments are `name`, `bases`, and `attributes`, as usual.
        """
        super().__init__(*args, **kwargs)
        cls._instances = {}

    def __call__(cls, *args, **kwargs):
        """
        The constructor/initializer require at least one argument since
        the first argument is the singleton key.

        If an instance was already created with the requested key, it is
        returned without being re-allocated or re-initialized.
        """
        if not args:
            raise ValueError(
                'Instance key must be first argument in constructor'
            )
        key = args[0]
        if key in cls._instances:
            instance = cls._instances[key]
        else:
            instance = super().__call__(*args, **kwargs)
            cls._instances[key] = instance
        return instance

    def reset(cls):
        """
        Resets/clears the class's registry so that new instances will be
        constructed for further calls to `cls`.
        """
        cls._instances.clear()


def shift_up(*iterables):
    """
    For a given sequence of iterables, return a sequence of iterables
    that has all the elements shifted as far left as possible.

    In other words, earlier iterables in the sequence are replaced by
    the next available iterable as they are exhausted.

    For example, given::

        a = [1]
        b = [2, 3]
        c = [4, 5, 6]
        d = [7, 8]

    ``a, b, c, d = map(list, shift_up(a, b, c, d))`` will rearrange
    the lists so that::

        a = [1, 3, 6]
        b = [2, 5]
        c = [4, 8]
        d = [7]

    The result is a list of lists.
    """
    # Get iterators
    iterables = [iter(iterable) for iterable in iterables]
    # Generate a list of tuples containing each row
    count = len(iterables)
    shifted = [[] for _ in range(count)]
    while count > 0:
        # Generate a row
        it = 0
        while it < count:
            try:
                shifted[it].append(next(iterables[it]))
            except StopIteration:
                # Remove expired iterable
                del iterables[it]
                count -= 1
            else:
                # Keep moving along row
                it += 1

    return shifted


def any_nsc(iterable):
    """
    A non-short-circuiting version of :py:func:`any`.

    Useful for situations where the side-effects of an iterator are
    useful, for example when the elements are produced by a callable in
    a generator that also logs the :py:obj:`True` elements.

    The default behavior ``any([]) == False`` is preserved.

    See https://stackoverflow.com/q/1790520/2988730 for source material.
    """
    return reduce(or_, (bool(x) for x in iterable), False)


def all_nsc(iterable):
    """
    A non-short-circuiting version of :py:func:`all`.

    Useful for situations where the side-effects of an iterator are
    useful, for example when the elements are produced by a callable in
    a generator that also logs the :py:obj:`False` elements.

    The default behavior ``all([]) == True`` is preserved.

    See https://stackoverflow.com/q/1790520/2988730 for source material.
    """
    return reduce(and_, (bool(x) for x in iterable), True)


class RangeBuilder:
    """
    Utility for keeping tracks of the bounds of multiple ranges, e.g.,
    for the limits of a plot.

    A new builder is uninitialized at first: `start` and `stop` are
    both `None`. To add a new range, call the `update` method. The
    `start` and `stop` properties get the cumulative range.

    Any types that support mutual comparison may be supplied to
    `update`. Types that support subtraction will enable the `span`
    property as well.
    """
    def __init__(self):
        """
        Construct a new, uninitialized range builder.
        """
        self._init = False

    def __repr__(self):
        """
        Create a string representation of this object.

        The result shows the range, and is not representative of an
        actual initializer call.
        """
        args = f'{self._start}, {self._stop}' \
                                        if self._init else '<uninitialized>'
        return f'{type(self).__name__}({args})'

    @property
    def init(self):
        """
       Inidcates whether the range has been initialized or not.
        """
        return self._init

    @property
    def start(self):
        """
        The minimum lower bound of any of the ranges encountered so
        far. `None` if not initialized with at least one range.
        """
        if self._init:
            return self._start

    @property
    def stop(self):
        """
        The maximum upper bound of any of the ranges encountered so
        far. `None` if not initialized.
        """
        if self._init:
            return self._stop

    @property
    def range(self):
        """
        A list containing `[start, stop]` for a range encompassing all
        the values encountered so far.  `None` if not initialized.
        """
        if self._init:
            return [self._start, self._stop]

    @property
    def span(self):
        """
        .. py:attribute:: span

           For types that support subtraction in addition to comparison,
           This is the difference between `stop` and `start`. `None` if
           not initialized.
        """
        if self._init:
            return self._stop - self._start

    def update(self, start, stop=None):
        """
        Add a range to the builder.

        The first call to this method initializes the builder.

        Parameters
        ----------
        start :
            Any object supporting `min` comparison against the prior
            entries.
        stop :
            Any object supporting `max` comparison against the prior
            entries. Optional.
        """
        if stop is None:
            stop = start
        if self._init:
            self._start = min(start, self._start)
            self._stop = max(stop, self._stop)
        else:
            self._start = start
            self._stop = stop
            self._init = True


class chained_getter:
    r"""
    Create a getter function similar to :py:func:`operator.attrgetter`
    that can traverse a chain of index and attribute lookups.

    The `spec` string is formatted very similarly to that of
    :py:func:`operator.attrgetter`, with the addition that indices may
    be specified as well, similarly to :py:func:`operator.itemgetter`.

    Indices can be integer, positional, or named arguments. To
    specify a positional index, simply leave the brackets blank and
    supply the value in `*args`. To specify a named argument, place the
    name in brackets and add the corresponding value to `**kwargs`.

    Parameters
    ----------
    spec : str
        A format string that determines a series of attribute and index
        lookups to perform.
    \*args :
        Positional arguments to fill in for empty bracket placeholders
        in the `spec`. The index is computed from the number of empty
        brackets only.
    \*\*kwargs :
        Named arguments to fill in for brackets with key identifiers
        in `spec`.

    Return
    ------
    A callable that can look up multiple nested attributes and indices,
    similarly to, but more generally than :py:func:`operator.attrgetter`
    and :py:func:`operator.itemgetter`.

    Examples
    --------
    The following will create a getter for the first element of
    attribute ``a``::

        chained_getter('a[0]')

    Attributes may be chained::

        chained_getter('a.b[1].c.d')

    For non-numerical indices, supply the index object as a positional
    argument by leaving empty brackets. When called on object ``x``,
    the following will look up ``x.a.b['key']``::

        chained_getter('a.b[]', 'key')

    Named arguments can also be used by placing an identifier name in
    the brackets and adding the corresponding value to ``kwargs``::

        chained_getter('a.b[name]', name='key')

    Indices can be chained just like attributes, and numbers can be
    specified as any other object. Named and positional indices can be
    mixed. Position index is computed from the number of empty brackets
    found. The following will index ``[3]['key']['index'][4][5][6]``::

        chained_getter('[][name1][][][name2][name3]',
                       3, 'index', 4, name1='key', name2=5, name3=6)
    """
    _index_pattern = re.compile(r'\[([^\]]*)\]')
    _attr_pattern = re.compile(r'[\.\w]*')

    def __init__(self, spec, *args, **kwargs):
        """
        Initialize a getter for the specified parameters.
        """
        def get_index(element):
            nonlocal pos_count

            if not element:
                idx = pos_count
                pos_count += 1
                return args[idx]
            if element.replace('-', '').isdigit():
                return int(element)
            return kwargs[element]

        def get_attribute(element):
            if not self._attr_pattern.fullmatch(element):
                raise SyntaxError('Invalid characer in attribute name')
            if any(attr[0].isdigit() for attr in element.split('.') if attr):
                raise SyntaxError('Attribute name may not start with a digit')
            return element

        elements = deque()
        pos_count = 0
        for i, element in enumerate(self._index_pattern.split(spec)):
            if i % 2:
                # Handle indices
                getter = itemgetter(get_index(element))
            elif element:
                # Handle non-empty attribute chains
                if i > 0:
                    if element[0] != '.':
                        raise SyntaxError('Attribute after index '
                                          'must start with .')
                    element = element[1:]
                getter = attrgetter(get_attribute(element))
            else:
                # Skip empty attributes (separating indices)
                continue
            elements.append(getter)

        self._elements = elements

    def __call__(self, target):
        """
        Extract nested attributes from `target` object.
        """
        for getter in self._elements:
            target = getter(target)
        return target
